# Module:         poFitsParse
# Copyright:      Paul Obermeier 2022 / paul@poSoft.de
# First Version:  2022 / 04 / 01
#
# Distributed under BSD license.
#
# Module for reading and writing of FITS image files with pure Tcl.

namespace eval poFitsParse {
    variable ns [namespace current]

    namespace ensemble create

    namespace export GetImageHeader
    namespace export GetImageData
    namespace export ReadImageHeader ReadImageFile

    #
    # Utility procedures.
    #

    proc _GetNativeByteOrder {} {
        if { $::tcl_platform(byteOrder) eq "littleEndian" } {
            return "Intel"
        } else {
            return "Motorola"
        }
    }

    proc _GetPixelSize { pixelType } {
        if { $pixelType eq "float" } {
            return 4
        } elseif { $pixelType eq "short" } {
            return 2
        } elseif { $pixelType eq "byte" } {
            return 1
        } else {
            error "Invalid PixelType value: $pixelType (must be byte or short)"
        }
    }

    proc _GetScanFormat { pixelType byteOrder } {
        if { $pixelType eq "float" } {
            if { $byteOrder eq "Intel" } {
                set scanFmt "r"
            } else {
                set scanFmt "R"
            }
        } elseif { $pixelType eq "short" } {
            if { $byteOrder eq "Intel" } {
                set scanFmt "su"
            } else {
                set scanFmt "Su"
            }
        } elseif { $pixelType eq "byte" } {
            set scanFmt "cu"
        } else {
            error "Invalid PixelType value: $pixelType (must be byte or short)"
        }
        return $scanFmt
    }

    proc _GetConversionParams { fitsObj } {
        set fmtString ""
        set pixelType ""

        # If the image has BZERO or BSCALE keywords in the header, fitsTcl will
        # do the appropriate thing with them automatically, but the datatype
        # returned will be floating point doubles (isn't FITS fun:)
        if { ( [catch { $fitsObj get keyword BZERO}]  == 0) ||
             ( [catch { $fitsObj get keyword BSCALE}] == 0) } {
            set fmtString "f*"
            set pixelType "float"
        } else {
            set imgType [$fitsObj info imgType]
            # Note, that 32 and 64 bit integer values are not handled by the RAW
            # image format, so the values are interpreted as shorts and may give
            # wrong results.
            switch -exact -- $imgType {
                  8 { set fmtString "c*" ; set pixelType "byte" }
                 16 { set fmtString "s*" ; set pixelType "short" }
                 32 { set fmtString "s*" ; set pixelType "short" }
                 64 { set fmtString "s*" ; set pixelType "short" }
                -32 { set fmtString "f*" ; set pixelType "float" }
                -64 { set fmtString "f*" ; set pixelType "float" }
            }
        }
        return [list $fmtString $pixelType]
    }

    #
    # Image header read/write procedures.
    #

    proc GetImageHeader { fitsObj { fileName "RawData" } } {
        # Get the header of a FITS image.
        #
        # fitsObj  - Identifier of the FITS image.
        # fileName - File name of the FITS image, if appropriate.
        #
        # Return the header information as a dictionary.
        #
        # See ReadImageHeader for a description of the dictionary elements.
        #
        # If the header contains invalid information, an error is thrown.
        #
        # See also: ReadImageHeader PutImageHeader

        set headerDict [dict create]

        lassign [$fitsObj info imgdim] width height numChan
        if { ! [info exists numChan] || $numChan eq "" } {
            set numChan 1
        }
        if { $numChan == 4 } {
            set numChan 3
        }
        lassign [_GetConversionParams $fitsObj] fmtString pixelType
        set byteOrder [_GetNativeByteOrder]

        # Store the values contained in the header.
        dict set headerDict Magic     "FITS"
        dict set headerDict Width     $width
        dict set headerDict Height    $height
        dict set headerDict NumChan   $numChan
        dict set headerDict ByteOrder $byteOrder
        dict set headerDict ScanOrder "BottomUp"
        dict set headerDict PixelType $pixelType

        # Calculate derived information from the header values.
        set pixelSize [_GetPixelSize  $pixelType]
        set scanFmt   [_GetScanFormat $pixelType $byteOrder]
        dict set headerDict PixelSize  $pixelSize
        dict set headerDict ScanFormat $scanFmt
        dict set headerDict NumPixel   [expr { $width * $height }]
        dict set headerDict NumByte    [expr { $width * $height * $numChan * $pixelSize }]
        dict set headerDict FileName   $fileName

        return $headerDict
    }

    proc ReadImageHeader { fitsImgFile } {
        # Read the header of a FITS image file.
        #
        # fitsImgFile - File name of the FITS image.
        #
        # Return the header information as a dictionary containing the following keys:
        # Magic Width Height NumChan ByteOrder ScanOrder PixelType
        # PixelSize ScanFormat NumPixel NumByte
        #
        # The first seven keys have identical names and values as listed in the file header.
        # PixelSize and ScanFormat are derived information values computed from the other keys.
        #
        # Magic      Magic string "FITS" for easy identification of file format.
        # Width      Width of the image in pixel (integer).
        # Height     Height of the image in pixel (integer).
        # NumChan    Number of channels contained in image (integer).
        # ByteOrder  "Motorola" for big-endian, "Intel" for little-endian architecture.
        # ScanOrder  "TopDown" or "BottomUp"
        # PixelType  "byte" for 1-byte unsigned integers, "short" for 2-byte unsigned integers,
        #            "float" for 4-byte single precision floating point values.
        # PixelSize  Pixel size in bytes: 1, 2 or 4.
        # ScanFormat Format string for Tcl "binary" command to read image data.
        # NumPixel   Number of pixels contained in image (integer).
        # NumByte    Number of bytes contained in image data (integer).
        #
        # If the header contains invalid information, an error is thrown.
        #
        # See also: GetImageHeader ReadImageFile

        set retVal [catch { fits open $fitsImgFile } fitsObj]
        if { $retVal != 0 } {
            error "ReadImageHeader: Cannot open file $fitsImgFile"
        }

        set headerDict [GetImageHeader $fitsObj [file tail $fitsImgFile]]
        $fitsObj close
        return $headerDict
    }

    #
    # Image data read/write procedures.
    #

    proc GetImageData { fitsObj headerDict dataDict } {
        upvar 1 $dataDict myDict

        set blankMapValue 0

        lassign [_GetConversionParams $fitsObj] fmtString pixelType
        set numChan [dict get $headerDict NumChan]
        set width   [dict get $headerDict Width]
        set height  [dict get $headerDict Height]

        # Retrieve the FITS image data as a Tcl list and convert into a byte array.
        set imgAsList [$fitsObj get image]

        # Check, if image has BLANK pixels, which are encoded as "NULL" in the Tcl list.
        # Replace those pixels with replacement value defined in blankMapValue.
        if { [lsearch -exact -ascii $imgAsList "NULL"] >= 0 } {
            set imgAsList [lmap a $imgAsList { expr { $a eq "NULL" ? $blankMapValue : $a }}]
        }

        if { $pixelType eq "short" } {
            # FITS uses signed short values, so map the pixel values to unsigned short for RAW images.
            set imgAsList [lmap p $imgAsList { expr { $p - 32768 }}]
        }

        if { $numChan == 3 || $numChan == 4 } {
            # FITS returns RGB images channel by channel, while RAW images need RGB.
            set chanSize [expr { $width * $height }]
            set sr 0
            set er [expr { $chanSize - 1 }]
            set sg [expr { $er + 1 }]
            set eg [expr { 2 * $chanSize - 1 }]
            set sb [expr { $eg + 1 }]
            set eb [expr { 3 * $chanSize - 1 }]
            foreach r [lrange $imgAsList $sr $er] \
                    g [lrange $imgAsList $sg $eg] \
                    b [lrange $imgAsList $sb $eb] {
                lappend rgbList $r $g $b
            }
            set imgAsList $rgbList
            set numChan 3
        }

        set imgAsByteArray [binary format $fmtString $imgAsList]

        dict set myDict Data $imgAsByteArray
    }

    proc ReadImageFile { fitsImgFile { optStr "" } } {
        # Read a FITS image file into a dictionary.
        #
        # fitsImgFile - File name of the FITS image.
        # optStr     - Option string for reading the FITS image.
        #              This is the value as returned by poImgType GetOptByFmt "fits" "read".
        #
        # Return the image data as a Tcl dictionary containing 2 keys: header and data.
        # Key "Data" contains the raw image data information as a binary string.
        # Key "Header" contains additional information about the image and is itself a dictionary.
        #
        # See ReadImageHeader for the description of the header dictionary.
        #
        # See also: ReadImageHeader WriteImageFile

        set retVal [catch { fits open $fitsImgFile } fitsObj]
        if { $retVal != 0 } {
            error "ReadImageFile: Cannot open file $fitsImgFile"
        }

        set imgDict [dict create]
        if { [poType IsImage $fitsImgFile "fits"] } {
            set headerDict [GetImageHeader $fitsObj [file tail $fitsImgFile]]
        }
        GetImageData $fitsObj $headerDict imgDict
        dict append imgDict Header $headerDict

        $fitsObj close
        return $imgDict
    }
}
